/*
 This usbio library is designed to communicate to an FPGA target via
 Cypress FX2 USB device. The library together with the FPGA local bus
 master (fx2.v) makes the FPGA local bus appear to the Linux host as
 a memory-mapped bus with 32-bit address and 32-bit data.
 
 2006-10-21: Created

 Copyright (c) 2006 Dimtel, Inc., All Rights Reserved

 $Id: usbio.c,v 1.1 2012/11/11 05:25:19 dim Exp $
 $Date: 2012/11/11 05:25:19 $

 $Author: dim $
 $Revision: 1.1 $
 $Log: usbio.c,v $
 Revision 1.1  2012/11/11 05:25:19  dim
 Moved from iGp12/src/usbio, updated to handle LLRF4 and iGp12

 Revision 1.2  2012/08/29 06:45:59  dim
 Added usbio_write_eeprom(), write_offsets command

 Revision 1.1.1.1  2010/10/05 16:42:44  dim
 Initial import of iGp12

 Revision 1.1.1.1  2009/06/02 20:26:02  dim
 Initial import

 Revision 1.2  2009/06/02 19:56:51  dim
 Full support for service functions (fw load, 8051 control, EEPROM) added

 Revision 1.1.1.1  2006/11/16 19:38:15  dim
 Initial import

*/

#include <stdio.h>
#include <error.h>
#include <errno.h>
#include <pthread.h>

#include "usbio.h"
#include "../include/eeprom.h"
#include "../include/usb_defs.h"
#include "../include/vendor_commands.h"
#include "../include/board_ids.h"

char *exnm = NULL;

struct usb_device *find_device(u_int16_t vid, u_int16_t pid, char *serial,
                              u_int16_t *bcdDevice)
{
  char ser_buf[1024];
  struct usb_bus *p;
  struct usb_device *q;
  usb_dev_handle *handle;

  p=usb_busses;
  while(p!=NULL)
  {
    q=p->devices;
    while(q!=NULL)
    {
      if((q->descriptor.idVendor == vid) && (q->descriptor.idProduct == pid))
      {
        if (serial)
        {
          if ((handle = usb_open(q)))
          {
            if (usb_get_string_simple(handle, q->descriptor.iSerialNumber, ser_buf, 1024))
              if (strcmp(ser_buf, serial) == 0)
              {
                usb_close(handle);
                return q;
              }
            usb_close(handle);
          }
        }
        else if (bcdDevice)
        {
          if (q->descriptor.bcdDevice == *bcdDevice)
            return q;
        }
        else
          return q;
      }
      q=q->next;
    }
    p=p->next;
  }
  return NULL;
}

usbio_context *usbio_open(int flag)
{
  return usbio_open_full(flag, NULL, NULL);
}

usbio_context *usbio_open_serial(int flag, char *serial)
{
  return usbio_open_full(flag, serial, NULL);
}

usbio_context *usbio_open_full(int flag, char *serial, unsigned short *bcd)
{
  int rc;
  usbio_context *ctxt;
  pthread_mutexattr_t attr;

  ctxt = (usbio_context *)malloc(sizeof(usbio_context));
  if (ctxt == NULL)
    return NULL;

  /* Set up the mutex */
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
  pthread_mutex_init(&ctxt->mutex, &attr);

  /* Get USB going */
  usb_init();
  usb_find_busses();
  usb_find_devices();

  /* Three possible VID/PID combinations: FX2, IGP12, LLRF4 */
  ctxt->device=find_device(FX2_DEFAULT_VID, FX2_DEFAULT_PID, serial, bcd);
  if(ctxt->device == NULL)
  {
    ctxt->device = find_device(IGP12_VID, IGP12_PID, serial, bcd);
    if (ctxt->device == NULL)
      ctxt->device = find_device(LLRF4_VID, LLRF4_PID, serial, bcd);
  }

  if(ctxt->device==NULL)
  {
#ifdef USBIO_DEBUG
    fprintf(stderr, 
    "%s: usbio_open(): devices 0x%04x:0x%04x, 0x%04x:0x%04x, and 0x%04x:0x%04x not found\n",
            exnm, FX2_DEFAULT_VID, FX2_DEFAULT_PID, IGP12_VID, IGP12_PID,
            LLRF4_VID, LLRF4_PID);
#endif
    free(ctxt);
    return NULL;
  }

  ctxt->handle = usb_open(ctxt->device);

  if (!flag)
  {
    if (usb_claim_interface(ctxt->handle, 0) < 0)
    {
#ifdef USBIO_DEBUG
      fprintf(stderr,"usbio_open: could not claim interface 0: %s\n", usb_strerror());
#endif
      usb_close(ctxt->handle);
      free(ctxt);
      return NULL;
    }
  
    /* Reset the FPGA state machine and FIFOs */
    rc = usbio_if_reset(ctxt);
    if (rc < 0)
    {
#ifdef USBIO_DEBUG
      fprintf(stderr,"%s: usbio_open(): usb_if_reset() failed: %s\n",
              exnm, usb_strerror());
#endif
      usbio_close(ctxt, 0);
      return NULL;
    }
    
    usb_set_altinterface(ctxt->handle, 1);
  }
  
  return ctxt;
}

void usbio_close(usbio_context *ctxt, int flag)
{
  if (!flag)
    usb_release_interface(ctxt->handle, 0);
  usb_close(ctxt->handle);
  free(ctxt);
}

unsigned int usbio_write(usbio_context *ctxt, unsigned int addr, 
                              unsigned int data)
{
  return usbio_block_write(ctxt, addr, &data, 1);
}

unsigned int usbio_block_write(usbio_context *ctxt, unsigned int addr, 
                      unsigned int *data, unsigned int count)
{
  unsigned int buf[USBIO_MAXLEN+2];
  unsigned int *ptr = data;
  unsigned int len, sz, rc;
  
#if USBIO_DEBUG==2
  local_thread_debug("usbio_block_write_lock", addr);
#endif
  if ((rc = pthread_mutex_lock(&ctxt->mutex)))
  {
    fprintf(stderr, "usbio_block_write(): pthread_mutex_lock() returned %d\n",
            rc);
    perror("usbio_block_write(): ");
  }
  
  while (count > 0)
  {
    len = MIN(count, USBIO_MAXLEN);
    count -= len;

    /* Form a transaction */    
    buf[0] = len & ~USBIO_M_RW;
    buf[1] = addr;
    memcpy(buf+2, ptr, len*sizeof(int));
    ptr  += len;
    addr += len;
    
    /* Send it out */
    sz = (len+2)*sizeof(int);
    rc = usb_bulk_write(ctxt->handle, USBIO_EP_WRITE, (char *)buf, 
                        sz, USBIO_TIMEOUT);
    if (rc != sz)
    {
      pthread_mutex_unlock(&ctxt->mutex);
#ifdef USBIO_DEBUG
      fprintf(stderr, "usbio_block_write: request for bulk write (0x%d) failed: %d\n",
              buf[1], rc);
#if USBIO_DEBUG==2
      local_thread_debug("usbio_block_write_unlocked_err", addr);
#endif
#endif
      return (rc-8)/4 + (unsigned int)(ptr - data);
    }
  }
  
  pthread_mutex_unlock(&ctxt->mutex);
#if USBIO_DEBUG==2
  local_thread_debug("usbio_block_write_unlocked", addr);
#endif
  return (unsigned int)(ptr - data);
}

unsigned int usbio_block_read(usbio_context *ctxt, unsigned int addr,
                     unsigned int *data, unsigned int count)
{
  unsigned int buf[2];
  unsigned int len, sz, rc;
  unsigned int *ptr = data;

#if USBIO_DEBUG==2
  local_thread_debug("usbio_block_read_lock", addr);
#endif
  if ((rc = pthread_mutex_lock(&ctxt->mutex)))
  {
    fprintf(stderr, "usbio_block_read(): pthread_mutex_lock() returned %d\n",
            rc);
    perror("usbio_block_read(): ");
  }
  
  while (count > 0)
  {
    len = MIN(count, USBIO_MAXLEN);
    count -= len;

    /* Form a transaction */    
    buf[0] = len | USBIO_M_RW;
    buf[1] = addr;
    addr += len;
    
    /* Send it out */
    sz = sizeof(buf);
    rc = usb_bulk_write(ctxt->handle, USBIO_EP_WRITE, (char *)buf, 
                        sz, USBIO_TIMEOUT);
    if (rc != sz)
    {
      pthread_mutex_unlock(&ctxt->mutex);
#ifdef USBIO_DEBUG
      fprintf(stderr, "usbio_block_read: request for bulk write (0x%x) failed: %d\n",
              buf[1], rc);
#if USBIO_DEBUG==2
      local_thread_debug("usbio_block_read_unlocked_err", addr);
#endif
#endif
      return (unsigned int)(ptr - data);
    }
    
    rc = usb_bulk_read(ctxt->handle, USBIO_EP_RDATA, (char *)ptr, 
                       len*sizeof(int), USBIO_TIMEOUT);
    if (rc != len*sizeof(unsigned int))
    {
      pthread_mutex_unlock(&ctxt->mutex);
#ifdef USBIO_DEBUG
      fprintf(stderr, "usbio_block_read: request for bulk read (0x%x) failed: %d\n",
              buf[1], rc);
#if USBIO_DEBUG==2
      local_thread_debug("usbio_block_read_unlocked_err", addr);
#endif
#endif
      return (unsigned int)(ptr - data + rc/sizeof(unsigned int));
    }
    ptr += len;
  }
  
  pthread_mutex_unlock(&ctxt->mutex);
#if USBIO_DEBUG==2
  local_thread_debug("usbio_block_read_unlocked", addr);
#endif
  return (unsigned int)(ptr - data);
}

unsigned int usbio_mult_write(usbio_context *ctxt, unsigned int *addr, 
                              unsigned int *data, int count)
{
/*  unsigned int buf[3];
  unsigned int bytes, idx, rc;
  struct usbdevfs_bulktransfer btf; */
  unsigned int idx;
  
  if (count*3*sizeof(int) > USBIO_MAXLEN)
    return 0;

  if (ctxt == NULL)
    return -1;
    
  for (idx = 0; idx < count; idx++)
    usbio_block_write(ctxt, addr[idx], data+idx, 1);
 
  return idx;

/*
  pthread_mutex_lock(&ctxt->mutex);
  
  btf.ep      = USBIO_EP_WRITE;
  btf.data    = (void *)buf;
  btf.timeout = USBIO_TIMEOUT;
  btf.len     = 3*sizeof(unsigned int);
  
  for (idx = 0; idx < count; idx++)
  {
    buf[0] = 1;
    buf[1] = addr[idx];
    buf[2] = data[idx];
    rc = ioctl(ctxt->fd, USBDEVFS_BULK, &btf);
  }
  
  pthread_mutex_unlock(&ctxt->mutex);
  
  return rc/3;
*/
}

static int usbio_vendor_cmd(usbio_context *ctxt, int bmRequest,
                            int Value, int Index, char *buf, int cnt)
{
  int rc;
  int bmRequestType = (bmRequest & USB_DIR_IN) | USB_TYPE_VENDOR;

  rc = usb_control_msg(ctxt->handle, bmRequestType, bmRequest,
                       Value, Index, (char *)buf, cnt, USBIO_TIMEOUT);
  return rc;
}

int usbio_if_reset(usbio_context *ctxt)
{
  int rc;

#if USBIO_DEBUG==2
  local_thread_debug("usbio_if_reset", 0);
#endif
  pthread_mutex_lock(&ctxt->mutex);
  /* Pulse FPGA RESET & reset FX2 FIFOs (handled by the firmware) */
  rc = usbio_vendor_cmd(ctxt, USB_REQ_VENDOR_FPGA_RESET, 0, 0, 
                          NULL, 0);
  pthread_mutex_unlock(&ctxt->mutex);
  return rc;
}

int usbio_i2c_write(usbio_context *ctxt, char addr, const void *buf, int len)
{
  if (len < 1 || len > EP0_MAX_PKT_SIZE)
    return -1;
  
  return usbio_vendor_cmd(ctxt, USB_REQ_VENDOR_I2C_WRITE, (int)addr, 0,
                          (char *)buf, len);
}

int usbio_i2c_read(usbio_context *ctxt, char addr, void *buf, int len)
{
  if (len < 1 || len > EP0_MAX_PKT_SIZE)
    return -1;
  
  return usbio_vendor_cmd(ctxt, USB_REQ_VENDOR_I2C_READ, (int)addr, 0,
                          (char *)buf, len);
}

int usbio_usberr_read(usbio_context *ctxt)
{
  char buf;
  int rc;
  
  rc = usbio_vendor_cmd(ctxt, USB_REQ_VENDOR_GET_STATUS, 0, 0, &buf, 1);
  
  if (rc != 1)
    return -1;
  else
    return (int)buf;
}

int usbio_write_sernum(usbio_context *ctxt, char *sernum)
{
  char *buf;
  int length = strlen(sernum), rc;
  
  if (length > 255)
    return -1;
  
  buf = (char *)malloc(length+5);
  
  if (buf == NULL)
    return -1;
  
  *buf     = (EEPROM_SERIAL >> 8) & 0xff;
  *(buf+1) = EEPROM_SERIAL & 0xff;
  *(buf+2) = EEPROM_SERIAL_HDR1;
  *(buf+3) = EEPROM_SERIAL_HDR2;
  *(buf+4) = length;
  memcpy(buf+5, sernum, length);
  
  rc = usbio_i2c_write(ctxt, I2C_EEPROM_ADDR, buf, length+5);
  free(buf);
  return rc;
}

int usbio_write_eeprom(usbio_context *ctxt, int addr, char *data, int length)
{
  char *buf;
  int rc;
  
  if (length > 32)
    return -1;
  
  buf = (char *)malloc(length+2);
  
  if (buf == NULL)
    return -1;
  
  *buf     = (addr >> 8) & 0xff;
  *(buf+1) = addr & 0xff;
  memcpy(buf+2, data, length);
  
  rc = usbio_i2c_write(ctxt, I2C_EEPROM_ADDR, buf, length+2);
  free(buf);
  return rc;
}

static int hex2nibble(char c)
{
  if (c >= '0' && c <= '9')
    return c - '0';
  if (c >= 'a' && c <= 'f')
    return c - 'a' + 10;
  if (c >= 'A' && c <= 'F')
    return c - 'A' + 10;
  
  /* Invalid */
  return -1;
}

int string2bytes(char *buf, char *string, int bufsize)
{
  int len = strlen(string), bptr = 0, sptr = 0, val;

  if (len % 2)
  {
    fprintf(stderr, "%s: data string must have even number of characters\n",
            exnm);
    return 0;
  }
  
  if (len/2 > bufsize)
  {
    fprintf(stderr, "%s: data length of %d exceeds maximum size of %d bytes\n",
            exnm, len>>1, bufsize);
    return 0;
  }
  
  while (*(string+sptr))
  {
    val = hex2nibble(*(string+sptr));
    if (val < 0)
    {
      fprintf(stderr, "%s: invalid character ""%c"" at location %d\n", 
              exnm, *(string+sptr), sptr);
      return 0;
    }
    *(buf+bptr) = (sptr & 1) ? *(buf+bptr) | val : val << 4;
    
    if (sptr++ & 1) bptr++;
  }
  return bptr;
}

int usbio_cpu_reset(usbio_context *ctxt, int value)
{
  char reg_value = value & 0xff;

  /* Cannot use usbio_vendor_cmd - single request for reading and writing */
  return usb_control_msg(ctxt->handle, USB_TYPE_VENDOR, USB_REQ_VENDOR_FW_LOAD,
                       0xe600, 0, &reg_value,  1, USBIO_TIMEOUT);
}

int usbio_readram(usbio_context *ctxt, int addr, int length, unsigned char **ram)
{
  unsigned char *p;
  int cnt;
  *ram = (unsigned char *)malloc(length);
  
  if (*ram == NULL)
    return -1;
  
  p = *ram;
  while (length)
  {
    cnt = usb_control_msg(ctxt->handle, USB_DIR_IN | USB_TYPE_VENDOR,
                       USB_REQ_VENDOR_FW_LOAD, addr, 0, (char *)p, 
                       MIN(USBIO_FW_CHUNK, length), USBIO_TIMEOUT);
    if (cnt < 0)
    {
      free(*ram);
      return -1;
    }
    else
    {
      length -= cnt;
      p      += cnt;
      addr   += cnt;
    }
  }
  return (p - *ram);
}

int usbio_writeram(usbio_context *ctxt, int addr, int length, unsigned char *data)
{
  unsigned char *p = data;
  int cnt;
  
  while (length)
  {
    cnt = usb_control_msg(ctxt->handle, USB_TYPE_VENDOR,
                       USB_REQ_VENDOR_FW_LOAD, addr, 0, (char *)p, 
                       MIN(USBIO_FW_CHUNK, length), USBIO_TIMEOUT);
    if (cnt < 0)
      return -1;
    else
    {
      length -= cnt;
      p      += cnt;
      addr   += cnt;
    }
  }
  return (p - data);
}

int usbio_fw_load(usbio_context *ctxt, char *file)
{
  /* Valid Intel HEX line is at most 521 character long. Add end of line
  symbols (<=2), and 0. Let there be some margin :) */
  char s[532], *p, eof = 0;
  unsigned char  byte, count, type, xsum, sum_val, data[255];
  unsigned short addr;
  int byte_count = 0, line = 0, rc;
  FILE *fid;

  fid = fopen(file, "r");
  if (fid == NULL)
  {
    error(0, errno, "usbio_fw_load(): file %s could not be opened for reading", file);
    return -1;
  }

  /* Start from halting 8051 */
  usbio_cpu_reset(ctxt, 1);

  while ((fgets(s, 532, fid) != NULL) && !eof)
  {
    p = s; line++;
    if (*p++ != ':')
    {
      error(0, 0, "usbio_fw_load(): invalid string in %s, start character %c, skipping",
            file, *s);
      continue;
    }
    /* Must have at least 10 characters for a valid record */
    if (strlen(p) < 10)
    {
      error(0, 0, "usbio_fw_load(): invalid record, line %d, skipping", line);
      continue;
    }
    count  = hex2nibble(*p++) << 4;
    count |= hex2nibble(*p++);
    addr   = hex2nibble(*p++) << 12;
    addr  |= hex2nibble(*p++) << 8;
    addr  |= hex2nibble(*p++) << 4;
    addr  |= hex2nibble(*p++);
    type   = hex2nibble(*p++) << 4;
    type  |= hex2nibble(*p++);
    
    /* Make sure we know what to do with this type */
    switch (type)
    {
      case 0: /* Read bytes, write to FX2 */
        /* Check length first */
        if (strlen(p) < (int)count * 2 + 2)
        {
          error(0, 0, "usbio_fw_load(): short data on line %d, %d bytes read, %d expected, aborting",
                line, (int)strlen(p), (int)count * 2);
          return -1;
        }
        xsum = count + (addr & 0xff) + (addr>>8) + type;
        for (byte = 0; byte < count; byte++)
        {
          data[byte]  = hex2nibble(*p++) << 4;
          data[byte] |= hex2nibble(*p++);
          xsum += data[byte];
        }
        
        /* Read checksum */
        sum_val  = hex2nibble(*p++) << 4;
        sum_val |= hex2nibble(*p++);
        if (((xsum+sum_val) & 0xff) != 0)
        {
          error(0, 0, "usbio_fw_load(): checksum error on line %d, aborting", line);
          return -1;
        }
        
        /* Write the data */
        if ((rc = usbio_writeram(ctxt, addr, count, data)) != count)
        {
          error(0, 0, "usbio_fw_load(): error writing data to FX2 (%d/%d), aborting\n",
                rc, count);
          return -1;
        }
        byte_count += count;
        break;
      case 1: /* EOF record, return */
        eof = 1;
        break;
      default:
        error(0, 0, "usbio_fw_load(): invalid record type %d\n", type);
    }
  }

  /* Release 8051 */
  usbio_cpu_reset(ctxt, 0);
  
  return byte_count;
}
